#include <mex.h>


/* MEX function ss=hgi(MAX,alpha,p,q,n,x,lambda)
 computes the truncated hypergeometric function pFq ^alpha(p;q;x*I_n)
 The sum is only over |kappa|<=MAX (and kappa<=lambda unless lambda=[])
 p and q are arrays, so hgi(30,a,[3 4],[5 6 7],3,0.5) is 
 2F3^a([3 4];[5 6 7];0.5*I_3) summed over all kappa with |kappa|<=30
 Uses the formula 
 J_lambda(x I_n)=x^|lambda| \prod_{(i,j)\in \lambda (n-(i-1)+alpha(j-1))

 by Plamen Koev, February 2004

  */


/* uses the C function hgi(int MAX, double alpha, int n, double* x, int nx, double* p, double* q,
                    int np, int nq, double *s)

Computes s[i]= np F nq ^alpha(p,q,x[i]I_n) for i=1,2,...,nx and the summation bring
|kappa|<=MAX

by Plamen Koev May 2004

*/


#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

typedef struct {
     double alpha, *x, *p, *q, *s, *z;
     int n, np, nq, nx, MAX, *l, *lc;
} glob;


void summation(int i, int ms, glob* a) {

  int j,m,ii;
  double zn,c,d,e,f,g;

  m=ms;
  if (i>1) if ((*a).l[i-1]<m) m=(*a).l[i-1];
  for (ii=1;ii<m+1;ii++) {
     (*a).l[i]=ii;
     (*a).lc[ii]+=1;        /* update conjugate partition */

	 zn=1;
     c=-(i-1)/(*a).alpha+(*a).l[i]-1;
     for (j=0;j<(*a).np;j++)  zn*=(*a).p[j]+c;
     for (j=0;j<(*a).nq;j++)  zn/=(*a).q[j]+c;

     d=(*a).l[i]*(*a).alpha-i;             /* update j_lambda  */
     for (j=1;j<(*a).l[i];j++) {
        e=d-j*(*a).alpha+(*a).lc[j];
        g=e+1;
        zn*=(g-(*a).alpha)*e/(g*(e+(*a).alpha));
     }
     for (j=1;j<i;j++) {
        f=(*a).l[j]*(*a).alpha-j-d;
        g=f+(*a).alpha;
        e=f*g;
        zn*=(e-f)/(g+e);
     }
	 zn*=((*a).n+(*a).alpha*c);
	 for (j=0;j<(*a).nx;j++) {
	    (*a).z[i*(*a).nx+j]*=(*a).x[j] * zn;
        (*a).s[j]+=(*a).z[i*(*a).nx+j];
	 }
	 if ((ms>ii)&&(i<(*a).n)) {
		 memcpy((*a).z+(i+1)*(*a).nx,(*a).z+i*(*a).nx,(*a).nx * sizeof(double));
		 summation(i+1,ms-ii,a);
	 }
  }
  (*a).l[i]=0;
  for (ii=1;ii<m+1;ii++) (*a).lc[ii]--;
}

void hgi(int MAX, double alpha, int n, double* x, int nx, double* p, double* q,
                    int np, int nq, double *s) {
  int i;
  glob *a;

  a=(glob*)mxMalloc(sizeof(glob));
  (*a).n=n;
  (*a).MAX=MAX;
  (*a).alpha=alpha;
  (*a).p = p;
  (*a).q = q;
  (*a).np = np;
  (*a).nq = nq;
  (*a).nx = nx;
  (*a).s=s;
  for (i=0;i<nx;i++) s[i]=1;
  (*a).x=x;

  (*a).l =(int*)mxMalloc(sizeof(int)*(MAX+1));
  memset((*a).l,0,sizeof(int)*(MAX+1));
  (*a).lc =(int*)mxMalloc(sizeof(int)*(MAX+1));
  memset((*a).lc,0,sizeof(int)*(MAX+1));
  (*a).z=(double*)mxMalloc(sizeof(double)*(MAX+1)*nx);
  memcpy((*a).z+nx,(*a).s,sizeof(double)*nx);    /* set to 1 also */

  summation(1,MAX,a);

  mxFree((*a).z);
  mxFree((*a).lc);
  mxFree((*a).l);
  mxFree(a);
}

void mexFunction( int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray*prhs[] )
{
   int MAX, n, nx, np, nq;
   double alpha, *x,*p, *q, *s;


  if (nrhs!=6)
    mexErrMsgTxt("Must have exactly six input arguments.");
  if (nlhs>1)
    mexErrMsgTxt("Too many output arguments.");
  
  
  if (!mxIsDouble(prhs[0]) || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2]) || !mxIsDouble(prhs[3]) || !mxIsDouble(prhs[4]) || !mxIsDouble(prhs[5]))
    mexErrMsgTxt("All inputs must be double arrays.");

  if (mxGetM(prhs[0])!=1 || mxGetN(prhs[0])!=1)
    mexErrMsgTxt("First input must be a scalar = max size of partition.");
  if (mxGetM(prhs[1])!=1 || mxGetN(prhs[1])!=1)
    mexErrMsgTxt("Second input must be a scalar = alpha.");
  if (mxGetM(prhs[2])>1 && mxGetN(prhs[2])>1)
    mexErrMsgTxt("Third input must be a vector = (a_1,...,a_p).");
  if (mxGetM(prhs[3])>1 && mxGetN(prhs[3])>1)
    mexErrMsgTxt("Forth input must be a vector = (b_1,...,b_q).");
  if (mxGetM(prhs[4])!=1 || mxGetN(prhs[4])!=1)
    mexErrMsgTxt("Fifth input must be a scalar = number of variables.");
  if (mxGetM(prhs[5])!=1 && mxGetN(prhs[5])!=1)
    mexErrMsgTxt("Sixth input must be a vector = x.");


  plhs[0]=mxDuplicateArray(prhs[5]);   /* output is the same size as input */
  s=mxGetPr(plhs[0]);

  x=mxGetPr(prhs[5]);
  nx=mxGetNumberOfElements(prhs[5]);  /* size of x */
  p=mxGetPr(prhs[2]);
  np=mxGetNumberOfElements(prhs[2]);  /* size of p */
  q=mxGetPr(prhs[3]);
  nq=mxGetNumberOfElements(prhs[3]);  /* size of q */


  MAX=*mxGetPr(prhs[0]);
  alpha=*mxGetPr(prhs[1]);
  n = *mxGetPr(prhs[4]);

  hgi(MAX,alpha,n,x,nx,p,q,np,nq,s);
}

